package dm_tests;
require Exporter;
@ISA    = qw(Exporter);
@EXPORT = qw(tests make_ifc_array);

#    Devmon: An SNMP data collector & page generator for the BigBrother &
#    Hobbit network monitoring systems
#    Copyright (C) 2005-2006  Eric Schwimmer
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.  Please see the file named
#    'COPYING' that was included with the distrubition for more details.


 # Modules
  use strict;
  use dm_config;
  use Math::BigInt::Calc;

 # Our global variable hash
  use vars qw(%g);
  *g = \%dm_config::g;

 # Global array and hash by descending priority/severity
   my %colors = ('red' => 4, 'yellow' => 3, 'clear' => 2, 'green' => 1); 
   my @color_order = sort {$colors{$b} <=> $colors{$a}} keys %colors;

 # Main test subroutine; parse data and feed it to the individual
 # test-specific subs
  sub tests {

    @{$g{'test_results'}} = (); # Our outgoing message queue  
    %{$g{'tmp_hist'}}     = (); # Temporary history hash

   # Timestamp
    $g{'testtime'} = time;

    do_log("Performing test logic",1);

   # Now go through each device and perform the test logic it needs
    for my $device (sort keys %{$g{'dev_data'}}) {
    
      my $oids = {};

     # Get template-specific variables
      my $vendor = $g{'dev_data'}{$device}{'vendor'};
      my $model  = $g{'dev_data'}{$device}{'model'};
      my $tests  = $g{'dev_data'}{$device}{'tests'};

     # If our tests = 'all', create a string with all the tests in it
      $tests = join ',', keys %{$g{'templates'}{$vendor}{$model}{'tests'}}
        if $tests eq 'all';

     # Copious, uncessary error checking, but whatever
      do_log("No vendor found for $device!", 0) and next if !defined $vendor;
      do_log("No model found for $device!", 0)  and next if !defined $model;
      do_log("No tests found for $device!", 0)  and next if !defined $tests;

     # Separate tests, perform individual test logic
      for my $test (split /,/, $tests) {

       # Make sure test template exists
        do_log("Could not find test template for '$test' on $device", 0)
          and next if !defined $g{'templates'}{$vendor}{$model}{'tests'}{$test};

       # Hash shortcut
        my $tmpl = \%{$g{'templates'}{$vendor}{$model}{'tests'}{$test}};

       # custom threshold pointer
        my $thr = \%{$g{'dev_data'}{$device}{'thresh'}{$test}};

       # Create our oids hash that will be populated by both snmp
       # data and transformed data.  This is done to keep me
       # from going insane when we start doing transforms
        oid_hash($oids, $device, $tmpl, $thr);

       # First transform any data and do threshold evaluations
        for my $oid (keys %$oids) {
         # Only transform if we dont already have values for this oid
          next if defined $oids->{$oid}{'val'};
          transform($device, $oids, $oid, $thr) if $oids->{$oid}{'transform'};
        }

       # Now crank out our message
        my $msg = render_msg($device, $tmpl, $test, $oids);

       # ... and add the message to our outgoing queue!
        push @{$g{'test_results'}}, $msg;

      }
    }

   # Now store our historical data, if any
    for my $device (keys %{$g{'tmp_hist'}}) {
      %{$g{'dev_hist'}{$device}} = %{$g{'tmp_hist'}{$device}};
    } 

   # Finish timestamp
    $g{'testtime'} = time - $g{'testtime'};
    do_log("Done with test logic",2);

  }



 # Create a oid hash ref that will eventually contain all gathered
 # data (snmp or transformed)
  sub oid_hash {

    my ($oids, $device, $tmpl, $thr) = @_;

   # Hash shortcuts
    my $snmp = \%{$g{'snmp_data'}{$device}};

   # For now we will copy the data from the template and snmp
    for my $oid (keys %{$tmpl->{'oids'}}) {

     # Dont hash an OID more than once
      next if defined $oids->{$oid}{'val'};

     # Put all the info we got on the oid in (sans transform data)
     # in the oids hash ref
     # First we do threshold data
      if(defined $tmpl->{'oids'}{$oid}{'thresh'}) {
        $oids->{$oid}{'thresh'} = $tmpl->{'oids'}{$oid}{'thresh'};
      }

     # Then exceptions
      if(defined $tmpl->{'oids'}{$oid}{'exceptions'}) {
        $oids->{$oid}{'except'} = $tmpl->{'oids'}{$oids}{'exceptions'};
      }

     # We dont have transform data yet, mark it for later and skip
      if(defined $tmpl->{'oids'}{$oid}{'trans_type'}) {
        $oids->{$oid}{'transform'} = 1;
        $oids->{$oid}{'trans_type'} = $tmpl->{'oids'}{$oid}{'trans_type'};
        $oids->{$oid}{'trans_data'} = $tmpl->{'oids'}{$oid}{'trans_data'};
        next;
      }

      my $num    = $tmpl->{'oids'}{$oid}{'number'};
      my $repeat = $tmpl->{'oids'}{$oid}{'repeat'}; 

      $oids->{$oid}{'transform'} = 0;
      $oids->{$oid}{'repeat'}    = $repeat;

      do_log("No SNMP data found for $oid on $device", 0) and next
        if !defined $num or !defined $snmp->{$num}
        or !defined $snmp->{$num}{'val'};

     # If this is a repeater, iterate through its leaves and assign values
      if($repeat) {
        for my $leaf (keys %{$snmp->{$num}{'val'}}) {
          # If we have a non-numeric leaf, make sure to keep track of this!
          # Store this as a type '2' repeater
          $oids->{$oid}{'repeat'} = 2 if $leaf !~ /^\d+$/;
          my $val  = $snmp->{$num}{'val'}{$leaf};
          my $time = $snmp->{$num}{'time'}{$leaf};
          $oids->{$oid}{'val'}{$leaf}  = $val;
          $oids->{$oid}{'time'}{$leaf} = $time;

        }

       # Apply thresholds
        apply_thresh_rep($oids, $thr, $oid);
      }

     # Otherwise its just a normal non-repeater
      else {
        my $val  = $snmp->{$num}{'val'};
        my $time = $snmp->{$num}{'time'};
        
        $oids->{$oid}{'val'}  = $val;
        $oids->{$oid}{'time'} = $time;

       # Apply threshold
        apply_thresh($oids, $thr, $oid);
        
      }
    }
    return $oids;
  }



 # Transform data values
  sub transform {
    my ($device, $oids, $oid, $thr) = @_;

   "$device:$oid DUP\n" and return if defined $oids->{$oid}{'val'};
   # Shortcut to our snmp data
    my $trans_type = $oids->{$oid}{'trans_type'};
    my $trans_data = $oids->{$oid}{'trans_data'};

   # Transform any oids that we depend on before we finish this one
    for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
      transform($device, $oids, $dep_oid) 
        if $oids->{$dep_oid}{'transform'}; 
    }

   # Make sure we inherit repeatability from previous types
    my $trans_sub = "trans_" . $trans_type;
    no strict 'refs';
    if(defined &$trans_sub) {
      &$trans_sub($device, $oids, $oid, $thr) if defined &$trans_sub;
    }
    else {
     # Theoretically we should never get here, but whatever
      do_log("Undefined transform type '$trans_type' found for $device", 0)
        and return;
    }
    use strict 'refs';

  }



 # Do data over time delta transformations
  sub trans_delta {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};


   # Hash shortcuts
    my $hist      = \%{$g{'dev_hist'}{$device}};
    my $hist_tmp  = \%{$g{'tmp_hist'}{$device}};
  
   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my ($dep_oid, $limit) = ($1, $2) 
      if $trans_data =~ /\{(.+)\}(\s*:\s*(\d+))?/;
    $limit = '' if !defined $limit;
    do_log("Bad delta transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_delta on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      LEAF: for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $this_data = $oids->{$dep_oid}{'val'}{$leaf};
        my $this_time = $oids->{$dep_oid}{'time'}{$leaf};
        my $regex = '^[-+]?\d+(\.\d+)?$';

       # Check data integrity
        if(!defined $this_data) {
          do_log("Cant find snmp data for leaf $leaf " .
                 "of $dep_oid on $device", 0);
          $oids->{$oid}{'val'}{$leaf}   = 'No data';
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = 'clear';
          next;
        }
        if($this_data !~ /$regex/) {
            do_log("Data type mismatch for leaf $leaf of $dep_oid on " .
                   "$device: '$this_data' does not match regex '$regex'", 0);
            $oids->{$oid}{'val'}{$leaf}   = 'Bad data';
            $oids->{$oid}{'time'}{$leaf}  = time;
            $oids->{$oid}{'color'}{$leaf} = 'clear';
        }

       # Check if we have history, return delta if so
        if(defined $hist->{$dep_oid}{'val'}{$leaf}
           and $this_time != $hist->{$dep_oid}{'time'}{$leaf}) {
          my $last_data = $hist->{$dep_oid}{'val'}{$leaf};
          my $last_time = $hist->{$dep_oid}{'time'}{$leaf};
          my $delta;

         # Check for counter wrap
          if($last_data > $this_data) {

           # Check for custom limit, first
            if($limit) {
              $delta = ($this_data + ($limit - $last_data)) /
                ($this_time - $last_time);
            }
           # If the value was less than 2^32, assume the counter was 32bit
            elsif($last_data < 4294967296) {
              $delta = ($this_data + (4294967296 - $last_data)) /
                ($this_time - $last_time);
            }
           # Otherwise the counter was 64bit...
            elsif($last_data < 18446744073709551616) {
              $delta = ($this_data + (18446744073709551616 - $last_data)) /
                ($this_time - $last_time);
            }

           # Otherwise something is seriously wrong
            else {
              do_log("Data type too large for leaf $leaf of " .
                     "$dep_oid on $device.", 0);
              $oids->{$oid}{'val'}{$leaf}   = 'Too large';
              $oids->{$oid}{'time'}{$leaf}  = time;
              $oids->{$oid}{'color'}{$leaf} = 'yellow';
              next LEAF;   
            }

            do_log("Counterwrap on $oid.$leaf on $device (this: $this_data " .
                   "last: $last_data delta: $delta",0) if $g{'debug'};
          }

         # Otherwise do normal delta calc
          else {
            $delta = ($this_data - $last_data) / 
              ($this_time - $last_time);
          }
          
         # Round delta to two decimal places
          $delta = sprintf "%.2f", $delta;

          $oids->{$oid}{'val'}{$leaf}   = $delta;
          $oids->{$oid}{'time'}{$leaf}  = time;
        }
  
       # No history; throw wait message
        else {
          $oids->{$oid}{'val'}{$leaf}   = 'wait';
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = 'green';
        }

       # Store history in temp hash
        $hist_tmp->{$dep_oid}{'val'}{$leaf}  = $this_data;
        $hist_tmp->{$dep_oid}{'time'}{$leaf} = $this_time;
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {

      return if validate_dep($device, $oids, $oid, $dep_oid, 
                             undef, '^[-+]?\d+(\.\d+)?$') != 1;

     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $this_data = $oids->{$dep_oid}{'val'};
      my $this_time = $oids->{$dep_oid}{'time'};

     # Check if we have history, return delta if so
      if(defined $hist->{$dep_oid}) {
        my $last_data = $hist->{$dep_oid}{'val'};
        my $last_time = $hist->{$dep_oid}{'time'};

        my $delta;

       # Check for counter wrap
        if($last_data > $this_data) {

         # If the value was less than 2^32, assume the counter was 32bit
          if($limit) {
            $delta = ($this_data + ($limit - $last_data)) /
              ($this_time - $last_time);
          }
          elsif($last_data < 4294967296) {
            $delta = ($this_data + (4294967296 - $last_data)) /
              ($this_time - $last_time);
          }
         # Otherwise the counter was 64bit...
          elsif($last_data < 18446744073709551616) {
            $delta = ($this_data + (18446744073709551616 - $last_data)) /
              ($this_time - $last_time);
          }

         # Otherwise something is seriously wrong
          else {
            do_log("Data type too large for $dep_oid on $device.", 0);
            $oids->{$oid}{'val'}   = 'Too large';
            $oids->{$oid}{'time'}  = time;
            $oids->{$oid}{'color'} = 'yellow';
            return;   
          }

          do_log("Counterwrap on $oid on $device (this: $this_data " .
                 "last: $last_data delta: $delta",0) if $g{'debug'};
        }

       # Otherwise do normal delta calc
        else {
          $delta = ($this_data - $last_data) / 
            ($this_time - $last_time);
        }

        $delta = sprintf "%.2f", $delta;
        $oids->{$oid}{'val'}  = $delta;
        $oids->{$oid}{'time'} = time;

       # Now apply our threshold to this data
        apply_thresh($oids, $thr, $oid);
      }

     # No history; throw wait message
      else {
        $oids->{$oid}{'val'}   = 'wait';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'green';
      }

     # Set temporary history data
      $hist_tmp->{$dep_oid}{'val'}  = $this_data;
      $hist_tmp->{$dep_oid}{'time'} = $this_time;
    }

    return;
  
  }



 # Do mathmatical translations
  sub trans_math {
    my ($device, $oids, $oid, $thr) = @_;
    my $expression = $oids->{$oid}{'trans_data'};
    my $repeat = 0;
    my $valid_non_reps = 1;
    my ($invalid_val, $invalid_time, $invalid_color);

   # Convert our math symbols to their perl equivalents
    $expression =~ s/x/\*/g;  # Multiplication
    $expression =~ s/\^/**/g; # Exponentiation

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    do_log("Bad transform data for $oid on $device", 0) and return
      if $expression !~ /\{.+}/;
 
   # Go through the expression and replace all non-repeater oids first
    for my $dep_oid ($expression =~ /\{(.+?)\}/g) {
     # Skip non-reps, but note that we have at least one
      if ($oids->{$dep_oid}{'repeat'}) {$repeat = 1; next} 

     # Check that this non-repeater is valid
      my $is_valid = validate_dep($device, $oids, $oid, $dep_oid, 
                                  undef, '^[-+]?\d+(\.\d+)?$');

     # If it isnt valid, copy our invalid values so we can assign them
     # To our leaves if we have repeating dependencies
      if($is_valid != 1) {
        $invalid_val   = $oids->{$oid}{'val'};
        $invalid_time  = $oids->{$oid}{'time'};
        $invalid_color = $oids->{$oid}{'color'};
        $valid_non_reps = $is_valid;
        next;
      }

     # Replace non-repeaters with their corresponding value
      my $val = $oids->{$dep_oid}{'val'};
      $expression =~ s/\{$dep_oid\}/$val/g;
    }

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat) {
      my $valid_non_reps = 1;

     # Okay, now all we have left in the expression is repeaters
     # Find the first one and use it as the key repeater
      my $primary_oid = $1 if $expression =~ /\{(.+?)}/;

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$primary_oid}{'repeat'};

     # Make sure our primary oid exists
      if(!defined $primary_oid) {
        do_log("Can't find primary oid for trans_math on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Cant determine primary oid";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }
      
     # Make sure we have leaf data for our primary oid
      elsif(!defined $oids->{$primary_oid}{'val'}) {
        do_log("Missing primary oid data for trans_math on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      MATHLEAF: for my $leaf (keys %{$oids->{$primary_oid}{'val'}}) {
       # If one or more of our non-repeaters was invalid, all leaves are too
        if($valid_non_reps != 1) {
          $oids->{$oid}{'val'}{$leaf}   = $invalid_val;
          $oids->{$oid}{'time'}{$leaf}  = $invalid_time;
          $oids->{$oid}{'color'}{$leaf} = $invalid_color;
          next;
        }

       # Copy our expression, then replace it, looking for errors first
        my $leaf_exp = $expression;
        my $regex    = '^[-+]?\d+(\.\d+)?$';
        for my $dep_oid ($expression =~ /\{(.+?)\}/g) {
          my $val = $oids->{$dep_oid}{'val'}{$leaf};
          if(!defined $val) {
            if($oids->{$dep_oid}{'transform'}) {
              do_log("Cant find translated data for leaf $leaf " .
                     "of $dep_oid on $device", 0);
            }
            else {
              do_log("Cant find snmp data for leaf $leaf " .
                     "of $dep_oid on $device", 0)
            }
            $oids->{$oid}{'val'}{$leaf}   = 'No data';
            $oids->{$oid}{'time'}{$leaf}  = time;
            $oids->{$oid}{'color'}{$leaf} = 'clear';
            next MATHLEAF;
          }

          elsif($oids->{$dep_oid}{'val'}{$leaf} eq 'wait') {
            $oids->{$oid}{'val'}{$leaf}   = 'wait';
            $oids->{$oid}{'time'}{$leaf}  = time;
            $oids->{$oid}{'color'}{$leaf} = 'green';
            next MATHLEAF;
          }

          elsif(defined $regex and $regex ne '' and $val !~ /$regex/) {
            do_log("Data type mismatch for leaf $leaf " .
                     "of $dep_oid on $device: '$val' does not match " .
                     "regex '$regex'", 0);
            my $val = $oids->{$dep_oid}{'val'}{$leaf};
            $oids->{$oid}{'val'}{$leaf}   = 'Bad data';
            $oids->{$oid}{'time'}{$leaf}  = time;
            $oids->{$oid}{'color'}{$leaf} = 'clear';
            next MATHLEAF;
          }

          $leaf_exp =~ s/\{$dep_oid\}/$val/g;
        }

        my $result;
        # Try to fix any divide by zero problems that might occur
        if($leaf_exp =~ /\s*\/\s*(0|0\.0+|\.0+)\s*(\)|$)/g) {
          # Its probably wrong to assume zero here, in the case
          # of a non-trivial mathmatical equation, but if someone
          # else wants to write code to parse this crap, they can do it :)
          $result = 0;
        }
        else {
          $result = eval($leaf_exp);
        }

        if($@) {
            do_log("Failed eval for TRANS_MATH on $oid.$leaf: $leaf_exp ($@)");
            $oids->{$oid}{'val'}{$leaf}   = 'Failed eval';
            $oids->{$oid}{'time'}{$leaf}  = time;
            $oids->{$oid}{'color'}{$leaf} = 'clear';
            next MATHLEAF;
        }

        $result = sprintf "%.2f", $result;
        $oids->{$oid}{'val'}{$leaf}  = $result;
        $oids->{$oid}{'time'}{$leaf} = time;
        
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

     # Make sure our dependencies if any are valid
      if($valid_non_reps != 1) {
        $oids->{$oid}{'val'}   = $invalid_val;
        $oids->{$oid}{'time'}  = $invalid_time;
        $oids->{$oid}{'color'} = $invalid_color;
        return;
      }

     # All of our non-reps were substituted earlier, so we can just eval

      my $result;
      # Try to fix any divide by zero problems that might occur
      if($expression =~ /\s*\/\s*(0|0\.0+|\.0+)\s*(\)|$)/) {
	# Its probably wrong to assume zero here, in the case
	# of a non-trivial mathmatical equation, but if someone
	# else wants to write code to parse this crap, they can do it :)
	$result = 0;
      }
      else {
	$result = eval $expression;
      }

      if($@) {
	  do_log("Failed eval for TRANS_MATH on $oid: $expression ($@)");
	  $oids->{$oid}{'val'}   = 'Failed eval';
	  $oids->{$oid}{'time'}  = time;
	  $oids->{$oid}{'color'} = 'clear';
	  next MATHLEAF;
      }

      $result = sprintf "%.2f", $result;
      $oids->{$oid}{'val'}  = $result;
      $oids->{$oid}{'time'} = time;
        
     # Now apply our threshold to this data
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Get substring of dependent oid
  sub trans_substr {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my ($dep_oid, $offset, $length) = ($1, $2, $3) 
      if $trans_data =~ /\{(.+)\}\s+(\d+)\s*(\d*)/;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_hex on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $string = $oids->{$dep_oid}{'val'}{$leaf};
        $oids->{$oid}{'val'}{$leaf}   = substr $string, $offset, $length;
        $oids->{$oid}{'time'}{$leaf}  = time;

       # Apply thresholds
	apply_thresh_rep($oids, $thr, $oid);
      }
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $string = $oids->{$dep_oid}{'val'};
      $oids->{$oid}{'val'}   = substr $string, $offset, $length;
      $oids->{$oid}{'time'}  = time;

     # Now apply our threshold to this data
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }


 # Set target oid to 1 if all source oids are non-zero, otherwise set to 0
  sub trans_and {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my $repeat_oid = '';
    my @dep_oids;
    while($trans_data =~ s/\{(.+?)\}//) {
      my $dep_oid = $1;
      $repeat_oid = $dep_oid 
        if $oids->{$dep_oid}{'repeat'} and $repeat_oid eq '';
      push @dep_oids, $dep_oid;
    }

    do_log("Bad transform data for $oid on $device", 0) and return
      if !@dep_oids;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat_oid ne '') {

     # Default values
      my $value = 1;
      my $color = 'green';
      my $msg   = $oids->{$repeat_oid}{'msg'};

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$repeat_oid}{'repeat'};

      AND_LEAF: for my $leaf (keys %{$oids->{$repeat_oid}{'val'}}) {
        my $value = 1;
        my $color = 'green';
        my $msg   = '';
        for my $dep_oid (@dep_oids) {
          my ($dep_val, $this_color, $this_msg);

          if($oids->{$dep_oid}{'repeat'}) {
            $dep_val    = $oids->{$dep_oid}{'val'}{$leaf};
            $this_color = $oids->{$dep_oid}{'color'}{$leaf};
            $this_msg   = $oids->{$dep_oid}{'msg'}{$leaf};
          }
          else {
            $dep_val    = $oids->{$dep_oid}{'val'};
            $this_color = $oids->{$dep_oid}{'color'};
            $this_msg   = $oids->{$dep_oid}{'msg'};
          }
       
         # Make sure we have leaf data for our dependent oid
          if(!defined $dep_val or !defined $this_color) {
            do_log("Source oid(s) undefined for trans_and on $oid", 0);
            $oids->{$oid}{'val'}{1}   = "Source oid(s) undefined";
            $oids->{$oid}{'time'}{1}  = time;
            $oids->{$oid}{'color'}{1} = 'red';
            next AND_LEAF;
          }
          
          $color = $this_color and $msg = $this_msg
            if $colors{$color} < $colors{$this_color};
          $value = 0 if !$dep_val;
        }

        $oids->{$oid}{'val'}{$leaf}   = $value;
        $oids->{$oid}{'time'}{$leaf}  = time;
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

     # Default values
      my $value = 1;
      my $color = 'green';
      my $msg   = '';

      for my $dep_oid (@dep_oids) {
        my $dep_val    = $oids->{$dep_oid}{'val'};
        my $this_color = $oids->{$dep_oid}{'color'};
        my $this_msg   = $oids->{$dep_oid}{'msg'};
     
       # Make sure we have leaf data for our dependent oid
        if(!defined $dep_val or !defined $this_color) {
          do_log("Source oid(s) undefined for trans_and on $oid", 0);
          $oids->{$oid}{'val'}{1}   = "Source oid(s) undefined";
          $oids->{$oid}{'time'}{1}  = time;
          $oids->{$oid}{'color'}{1} = 'red';
          next AND_LEAF;
        }
        
        $color = $this_color and $msg = $this_msg 
          if $colors{$color} < $colors{$this_color};
        $value = 0 if !$dep_val;
      }

      $oids->{$oid}{'val'}   = $value;
      $oids->{$oid}{'time'}  = time;

     # Apply thresholds
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Set target oid to 1 if ANY source oids are non-zero, otherwise set to 0
  sub trans_or {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my $repeat_oid = '';
    my @dep_oids;
    while($trans_data =~ s/\{(.+?)\}//) {
      my $dep_oid = $1;
      $repeat_oid = $dep_oid 
        if $oids->{$dep_oid}{'repeat'} and $repeat_oid eq '';
      push @dep_oids, $dep_oid;
    }

    do_log("Bad transform data for $oid on $device", 0) and return
      if !@dep_oids;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat_oid ne '') {

     # Default values
      my $value = 0;
      my $color = 'green';
      my $msg   = $oids->{$repeat_oid}{'msg'};

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$repeat_oid}{'repeat'};

      AND_LEAF: for my $leaf (keys %{$oids->{$repeat_oid}{'val'}}) {
        my $value = 1;
        my $color = 'green';
        my $msg   = '';
        for my $dep_oid (@dep_oids) {
          my ($dep_val, $this_color, $this_msg);

          if($oids->{$dep_oid}{'repeat'}) {
            $dep_val    = $oids->{$dep_oid}{'val'}{$leaf};
            $this_color = $oids->{$dep_oid}{'color'}{$leaf};
            $this_msg   = $oids->{$dep_oid}{'msg'}{$leaf};
          }
          else {
            $dep_val    = $oids->{$dep_oid}{'val'};
            $this_color = $oids->{$dep_oid}{'color'};
            $this_msg   = $oids->{$dep_oid}{'msg'};
          }
       
         # Make sure we have leaf data for our dependent oid
          if(!defined $dep_val or !defined $this_color) {
            do_log("Source oid(s) undefined for trans_or on $oid", 0);
            $oids->{$oid}{'val'}{1}   = "Source oid(s) undefined";
            $oids->{$oid}{'time'}{1}  = time;
            $oids->{$oid}{'color'}{1} = 'red';
            next AND_LEAF;
          }
          
          $color = $this_color and $msg = $this_msg
            if $colors{$color} < $colors{$this_color};
          $value = 1 if $dep_val;
        }

        $oids->{$oid}{'val'}{$leaf}   = $value;
        $oids->{$oid}{'time'}{$leaf}  = time;
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

     # Default values
      my $value = 0;
      my $color = 'green';
      my $msg   = '';

      for my $dep_oid (@dep_oids) {
        my $dep_val    = $oids->{$dep_oid}{'val'};
        my $this_color = $oids->{$dep_oid}{'color'};
        my $this_msg   = $oids->{$dep_oid}{'msg'};
     
       # Make sure we have leaf data for our dependent oid
        if(!defined $dep_val or !defined $this_color) {
          do_log("Source oid(s) undefined for trans_or on $oid", 0);
          $oids->{$oid}{'val'}{1}   = "Source oid(s) undefined";
          $oids->{$oid}{'time'}{1}  = time;
          $oids->{$oid}{'color'}{1} = 'red';
          next AND_LEAF;
        }
        
        $color = $this_color and $msg = $this_msg 
          if $colors{$color} < $colors{$this_color};
        $value = 1 if $dep_val;
      }

      $oids->{$oid}{'val'}   = $value;
      $oids->{$oid}{'time'}  = time;

     # Apply thresholds
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Translate hex or octal data into decimal
  sub trans_unpack {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my ($dep_oid, $type) = ($1, $2) if $trans_data =~ /\{(.+)\}\s+(\S+)/;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_unpack on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $packed = $oids->{$dep_oid}{'val'}{$leaf};
        $oids->{$oid}{'val'}{$leaf}   = unpack "$type*", $packed;
        $oids->{$oid}{'time'}{$leaf}  = time;
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $packed = $oids->{$dep_oid}{'val'};
      $oids->{$oid}{'val'}   = unpack "$type*", $packed;
      $oids->{$oid}{'time'}  = time;

     # Apply thresholds
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Translate hex or octal data into decimal
  sub trans_convert {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my ($dep_oid, $type, $pad) = ($1, $2, $3) 
      if $trans_data =~ /\{(.+)\}\s+(hex|oct)\s*(\d*)/i;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   $type = lc $type;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_hex on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $val = $oids->{$dep_oid}{'val'}{$leaf};
        my $int;
        if($type eq 'oct') {$int = oct $val}
        else {              $int = hex $val}
        $int = sprintf '%'."$pad.$pad".'d', $int if $pad ne '';
        $oids->{$oid}{'val'}{$leaf}   = $int;
        $oids->{$oid}{'time'}{$leaf}  = time;
      }

     # Apply thresholds
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $val = $oids->{$dep_oid}{'val'};
      my $int;
      if($type eq 'oct') {$int = oct $val}
      else {              $int = hex $val}
      $int = sprintf '%'."$pad.$pad".'d', $int if $pad ne '';
      $oids->{$oid}{'val'}   = $int;
      $oids->{$oid}{'time'}  = time;

     # Apply thresholds
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Get the best condition of one or more oids
  sub trans_best {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};
    my $repeat = 0;
    my $valid_non_reps = 1;
    my ($invalid_val, $invalid_time, $invalid_color);

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    do_log("Bad transform data for $oid on $device", 0) and return
      if $trans_data !~ /\{.+}/;
 
    for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
     # Skip non-reps, but note that we have at least one
      if ($oids->{$dep_oid}{'repeat'}) {$repeat = 1; next} 

     # Check that this non-repeater is valid
      my $is_valid = validate_dep($device, $oids, $oid, $dep_oid, 0);

     # If it isnt valid, copy our invalid values so we can assign them
     # To our leaves if we have repeating dependencies
      if($is_valid != 1) {
        $invalid_val   = $oids->{$oid}{'val'};
        $invalid_time  = $oids->{$oid}{'time'};
        $invalid_color = $oids->{$oid}{'color'};
        $valid_non_reps = $is_valid;
        next;
      }
    }

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat) {
      my $valid_non_reps = 1;
      my ($invalid_val, $invalid_time, $invalid_color);

      my $primary_oid = $1 if $trans_data =~ /\{(.+?)}/;

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$primary_oid}{'repeat'};

     # Make sure we have leaf data for our primary oid
      if(!defined $oids->{$primary_oid}{'val'}) {
        do_log("Missing repeater data for trans_best on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      BESTLEAF: for my $leaf (keys %{$oids->{$primary_oid}{'val'}}) {

       # If one or more of our non-repeaters was invalid, all leaves are too
        if($valid_non_reps != 1) {
          $oids->{$oid}{'val'}{$leaf}   = $invalid_val;
          $oids->{$oid}{'time'}{$leaf}  = $invalid_time;
          $oids->{$oid}{'color'}{$leaf} = $invalid_color;
          next;
        }

        for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
          my $result = validate_dep($device, $oids, $oid, $dep_oid,
            $leaf);
          if($result != 1) {
            do_log("Dependency error: $device, $oid, $dep_oid, $leaf",1);
            next;
          }

          if(!defined $oids->{$oid}{'color'}{$leaf} or
             $colors{$oids->{$dep_oid}{'color'}{$leaf}} <=
             $colors{$oids->{$oid}{'color'}{$leaf}}) {
            $oids->{$oid}{'val'}{$leaf} = $oids->{$dep_oid}{'val'}{$leaf};
            $oids->{$oid}{'color'}{$leaf} = $oids->{$dep_oid}{'color'}{$leaf};
            $oids->{$oid}{'time'}{$leaf}  = $oids->{$dep_oid}{'time'}{$leaf};
            $oids->{$oid}{'msg'}{$leaf}   = $oids->{$dep_oid}{'msg'}{$leaf};
          }
        }
      }
    }

   # Otherwise we are a single entry datum
    else {
      for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
        next if validate_dep($device, $oids, $oid, $dep_oid, 0) != 1;

       # Set our repeater type
        $oids->{$oid}{'repeat'} = 0;

        if(!defined $oids->{$oid}{'color'} or
           $colors{$oids->{$dep_oid}{'color'}} <=
           $colors{$oids->{$oid}{'color'}}) {
          $oids->{$oid}{'val'}   = $oids->{$dep_oid}{'val'};
          $oids->{$oid}{'color'} = $oids->{$dep_oid}{'color'};
          $oids->{$oid}{'time'}  = $oids->{$dep_oid}{'time'};
          $oids->{$oid}{'msg'}   = $oids->{$dep_oid}{'msg'};
        }
      }
    }

    return;
  }



 # Get the worst condition of one or more oids
  sub trans_worst {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};
    my $repeat = 0;
    my $valid_non_reps = 1;
    my ($invalid_val, $invalid_time, $invalid_color);

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    do_log("Bad transform data for $oid on $device", 0) and return
      if $trans_data !~ /\{.+}/;
 
    for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
     # Skip non-reps, but note that we have at least one
      if ($oids->{$dep_oid}{'repeat'}) {$repeat = 1; next} 

     # Check that this non-repeater is valid
      my $is_valid = validate_dep($device, $oids, $oid, $dep_oid, 
                                  undef, '^[-+]?\d+(\.\d+)?$');

     # If it isnt valid, copy our invalid values so we can assign them
     # To our leaves if we have repeating dependencies
      if($is_valid != 1) {
        $invalid_val   = $oids->{$oid}{'val'};
        $invalid_time  = $oids->{$oid}{'time'};
        $invalid_color = $oids->{$oid}{'color'};
        $valid_non_reps = $is_valid;
        next;
      }
    }

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat) {
      my $valid_non_reps = 1;
      my ($invalid_val, $invalid_time, $invalid_color);

      my $primary_oid = $1 if $trans_data =~ /\{(.*)\}/;

     # Set our repeater type
      $oids->{$oid}{'repeat'} =  $oids->{$primary_oid}{'repeat'};

     # Make sure we have leaf data for our primary oid
      if(!defined $oids->{$primary_oid}{'val'}) {
        do_log("Missing repeater data for trans_worst on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      WORSTLEAF: for my $leaf (keys %{$oids->{$primary_oid}{'val'}}) {

       # If one or more of our non-repeaters was invalid, all leeaves are too
        if($valid_non_reps != 1) {
          $oids->{$oid}{'val'}{$leaf}   = $invalid_val;
          $oids->{$oid}{'time'}{$leaf}  = $invalid_time;
          $oids->{$oid}{'color'}{$leaf} = $invalid_color;
          next;
        }

        for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
          next WORSTLEAF 
          if validate_dep($device, $oids, $oid, $dep_oid, 
                          $leaf, '^[-+]?\d+(\.\d+)?$') != 1;

          if(!defined $oids->{$oid}{'color'}{$leaf} or
             $colors{$oids->{$dep_oid}{'color'}{$leaf}} >=
             $colors{$oids->{$oid}{'color'}{$leaf}}) {
            $oids->{$oid}{'color'}{$leaf} = $oids->{$dep_oid}{'color'}{$leaf};
            $oids->{$oid}{'time'}{$leaf}  = $oids->{$dep_oid}{'time'}{$leaf};
            $oids->{$oid}{'msg'}{$leaf}   = $oids->{$dep_oid}{'msg'}{$leaf};
          }
        }
      }
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      for my $dep_oid ($trans_data =~ /\{(.+?)\}/g) {
        next WORSTLEAF 
        if validate_dep($device, $oids, $oid, $dep_oid, undef, 
          '^[-+]?\d+(\.\d+)?$') != 1;

        if(!defined $oids->{$oid}{'color'} or
           $colors{$oids->{$dep_oid}{'color'}} >=
           $colors{$oids->{$oid}{'color'}}) {
          $oids->{$oid}{'color'} = $oids->{$dep_oid}{'color'};
          $oids->{$oid}{'time'}  = $oids->{$dep_oid}{'time'};
          $oids->{$oid}{'msg'}   = $oids->{$dep_oid}{'msg'};
        }
      }
    }

    return;
  }

 # Return an time elapsed, seperated into years, weeks, days, hours, and seconds
  sub trans_elapsed {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my $dep_oid = $1 if $trans_data =~ /\{(.+)\}/;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_elapsed on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {

        my $secs = $oids->{$dep_oid}{'val'}{$leaf};
        $oids->{$oid}{'val'}{$leaf}   = elapsed($secs);
        $oids->{$oid}{'time'}{$leaf}  = time;
        $oids->{$oid}{'color'}{$leaf} = $oids->{$dep_oid}{'color'}{$leaf};
        $oids->{$oid}{'msg'}{$leaf}   = $oids->{$dep_oid}{'msg'}{$leaf};
       
      }
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $secs = $oids->{$dep_oid}{'val'};
      $oids->{$oid}{'val'}   = elapsed($secs);
      $oids->{$oid}{'time'}  = time;
      $oids->{$oid}{'color'} = $oids->{$dep_oid}{'color'};
      $oids->{$oid}{'msg'}   = $oids->{$dep_oid}{'msg'};
    }

    return;
  }


 # Return an time elapsed, seperated into years, weeks, days, hours, and seconds
  sub trans_date {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my $dep_oid = $1 if $trans_data =~ /\{(.+)\}/;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_elapsed on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        next;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {

        my $secs = $oids->{$dep_oid}{'val'}{$leaf};
        $oids->{$oid}{'val'}{$leaf}   = date($secs);
        $oids->{$oid}{'time'}{$leaf}  = time;
        $oids->{$oid}{'color'}{$leaf} = $oids->{$dep_oid}{'color'}{$leaf};
        $oids->{$oid}{'msg'}{$leaf}   = $oids->{$dep_oid}{'msg'}{$leaf};
       
      }
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $secs = $oids->{$dep_oid}{'val'};
      $oids->{$oid}{'val'}   = date($secs);
      $oids->{$oid}{'time'}  = time;
      $oids->{$oid}{'color'} = $oids->{$dep_oid}{'color'};
      $oids->{$oid}{'msg'}   = $oids->{$dep_oid}{'msg'};
    }

    return;
  }



 # Give a bps value, return a speed measured in the maximum unit-size possible
  sub trans_speed {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};

   # Speed hash
    my %speeds = ( 1             => 'bps',
                   1000          => 'Kbps',
                   1000000       => 'Mbps',
                   1000000000    => 'Gbps',
                   1000000000000 => 'Tbps');


   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    my $dep_oid = $1 if $trans_data =~ /\{(.+)\}/;
    do_log("Bad transform data for $dep_oid on $device", 0) and return
      if !defined $dep_oid;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_elapsed on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }
      
      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {

        my $bps = $oids->{$dep_oid}{'val'}{$leaf};

       # If this is a delta, wait for it
        if($bps eq 'wait') {
          $oids->{$oid}{'val'}{$leaf}   = 'wait';
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = 'green';
          next;
        }
          
       # Make sure our given speed is numeric
        elsif($bps !~ /^-?\d+(\.\d+)?$/) {
          $oids->{$oid}{'val'}{$leaf}   = 'Non-numeric speed';
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = 'yellow';
          next;
        }

        $bps = int $bps;

        for my $speed (sort {$b <=> $a} keys %speeds) {
          next if ($speed > $bps and $speed != 1);
          my $unit = $speeds{$speed};
          my $new_speed = $bps / $speed;

         # Measure to 2 decimal places
          $new_speed = sprintf "%.2f", $new_speed if $new_speed =~ /\./;

         # Add the unit measurement
          $new_speed .= " $unit";

          $oids->{$oid}{'val'}{$leaf}   = $new_speed;
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = $oids->{$dep_oid}{'color'}{$leaf};
          $oids->{$oid}{'msg'}{$leaf}   = $oids->{$dep_oid}{'msg'}{$leaf};
          last;
        }
      }
    }

   # Otherwise we are a single entry datum
    else {
     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      my $bps = $oids->{$dep_oid}{'val'};

     # If this is a delta, wait for it
      if($bps eq 'wait') {
        $oids->{$oid}{'val'}   = 'wait';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'green';
        next;
      }

     # Make sure our given speed is numeric
      elsif($bps =~ /\D/) {
        $oids->{$oid}{'val'}   = 'Non-numeric speed';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'yellow';
      }

      for my $speed (keys %speeds) {
        next if ($speed > $bps and $speed != 1);
        my $unit = $speeds{$speed};
        my $new_speed = $bps / $speed;

       # Measure to 2 decimal places
        $new_speed = sprintf "%.2f", $new_speed;

       # Add the unit measurement
        $new_speed .= " $unit";

        $oids->{$oid}{'val'}   = $new_speed;
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = $oids->{$dep_oid}{'color'};
        $oids->{$oid}{'msg'}   = $oids->{$dep_oid}{'msg'};
        last;
      }
    }

    return;
  }



 # Do a c-style case compare (but with more featureful variable matching)
 # against an array of if->then value pairs
  sub trans_switch {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};
    my %cases;

    my ($dep_oid, $switch_data) = ($1, $2) if
      $trans_data =~ /\{(.+?)}\s*(.+)/;

   # Populate our case hash
    my $case_num = 0;
    my $default;
    for my $val_pair (split /\s*,\s*/, $switch_data) {
      my ($if, $then) = ($1, $2) 
        if $val_pair =~ /^\s*(\S+.*?)\s*=\s*(.+?)\s*$/;
      do_log("Failed to extract value pair for case transform ($val_pair)", 0)
        and return if !defined $if;
      my $type = '';
      if($if =~ /^\d+$/) {$type = 'num'}
      elsif($if =~ /^>\s*([+-]?\d+(?:\.\d+)?)$/)  {$type = 'gt'; $if = $1}
      elsif($if =~ /^>=\s*([+-]?\d+(?:\.\d+)?)$/) {$type = 'gte'; $if = $1}
      elsif($if =~ /^<\s*([+-]?\d+(?:\.\d+)?)$/)  {$type = 'lt'; $if = $1}
      elsif($if =~ /^<=\s*([+-]?\d+(?:\.\d+)?)$/) {$type = 'lte'; $if = $1}
      elsif($if =~ /^([+-]?\d+(?:\.\d+)?)\s*-\s*([+-]?\d+(?:\.\d+)?)$/) {
        $type = 'rng'; $if = "$1-$2";
      }
      elsif($if =~ /^'(.+)'$/) {$type = 'str'; $if = $1}
      elsif($if =~ /^"(.+)"$/) {$type = 'reg'; $if = $1}
      elsif($if =~ /^default$/i) {$default = $then; next}
      
      do_log("Could not determine case type for string '$if'", 0)
         and return if !defined $if;
      $cases{++$case_num}{'if'} = $if;
      $cases{$case_num}{'type'} = $type;
      $cases{$case_num}{'then'} = $then;
    }
    my @case_nums = sort {$a <=> $b} keys %cases;

   # Make sure we have a default value
    $default = 'Unknown' if !defined $default;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)

    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_switch on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $val = $oids->{$dep_oid}{'val'}{$leaf};

        my $num;
        for my $case (@case_nums) {
          my $if   = $cases{$case}{'if'};
          my $type = $cases{$case}{'type'};
          if($type eq 'num') { $num = $case and last if $val == $if; next}
          if($type eq 'gt')  { $num = $case and last if $val > $if; next}
          if($type eq 'gte') { $num = $case and last if $val >= $if; next}
          if($type eq 'lt')  { $num = $case and last if $val < $if; next}
          if($type eq 'lte') { $num = $case and last if $val <= $if; next}
          if($type eq 'rng') { 
            my ($ll,$ul) = split /-/, $if;
            $num = $case and last if $val >= $ll and $val <= $ul; 
            next;
          }
          if($type eq 'str') { $num = $case and last if $val eq $if; next}
          if($type eq 'reg') { $num = $case and last if $val =~ /$if/; next}
        }
          
        my $then;
        if(defined $num) { $then = $cases{$num}{'then'} }
        else { $then = $default }
        while($then =~ /\{(\S+?)\}/) {
          my $then_oid = $1;
          my $then_oid_val;
          if($oids->{$then_oid}{'repeat'}) {
            $then_oid_val = $oids->{$then_oid}{'val'}{$leaf};
          }
          else {$then_oid_val = $oids->{$then_oid}{'val'}}
          if(!defined $then_oid_val) {
            do_log("Missing repeater data for trans_switch on $oid", 0);
            $then_oid_val = 'Undefined';
          }
          $then =~ s/\{$then_oid\}/$then_oid_val/g;
        }
          
        $oids->{$oid}{'val'}{$leaf}   = $then;
        $oids->{$oid}{'time'}{$leaf}  = time;

      }

      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
      my $val = $oids->{$dep_oid}{'val'};

     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      if(!defined $val) {
        do_log("No snmp data found for $dep_oid on $device", 0);
        $oids->{$oid}{'val'}   = "No SNMP data ";
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'clear';
        return;
      }

      my $num;
      for my $case (@case_nums) {
	my $if   = $cases{$case}{'if'};
	my $type = $cases{$case}{'type'};
	if($type eq 'num') { $num = $case and last if $val == $if; next}
	if($type eq 'gt')  { $num = $case and last if $val > $if; next}
	if($type eq 'gte') { $num = $case and last if $val >= $if; next}
	if($type eq 'lt')  { $num = $case and last if $val < $if; next}
	if($type eq 'lte') { $num = $case and last if $val <= $if; next}
	if($type eq 'rng') { 
	  my ($ll,$ul) = split /-/, $if;
	  $num = $case and last if $val >= $ll and $val <= $ul; 
	  next;
	}
	if($type eq 'str') { $num = $case and last if $val eq $if; next}
	if($type eq 'reg') { $num = $case and last if $val =~ /$if/; next}
      }
	
      my $then;
      if(defined $num) { $then = $cases{$num}{'then'} }
      else { $then = $default }

      while($then =~ /\{(\S+?)\}/) {
	my $then_oid = $1;
	my $then_oid_val;
	if($oids->{$then_oid}{'repeat'}) {
          do_log("Cant switch to a repeater OID when using a non-repeater" .
            "source OID for trans_tswitch on $oid", 0);
          $then_oid_val = 'Undefined';
	}
	else {$then_oid_val = $oids->{$then_oid}{'val'}}
	if(!defined $then_oid_val) {
	  do_log("Missing repeater data for trans_switch on $oid", 0);
	  $then_oid_val = 'Undefined';
	}
	$then =~ s/\{$then_oid\}/$then_oid_val/g;
      }
	
      $oids->{$oid}{'val'}   = $then;
      $oids->{$oid}{'time'}  = time;
      
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }


 # Do a c-style case compare (but with more featureful variable matching)
 # against an array of if->then value pairs, and also threshold inheritance
  sub trans_tswitch {
    my ($device, $oids, $oid, $thr) = @_;

    my $trans_data = $oids->{$oid}{'trans_data'};
    my %cases;

    my ($dep_oid, $switch_data) = ($1, $2) if
      $trans_data =~ /\{(.+?)}\s*(.+)/;

   # Populate our case hash
    my $case_num = 0;
    my $default;
    for my $val_pair (split /\s*,\s*/, $switch_data) {
      my ($if, $then) = ($1, $2) 
        if $val_pair =~ /^\s*(\S+.*?)\s*=\s*(.+?)\s*$/;
      do_log("Failed to extract value pair for case transform ($val_pair)", 0)
        and return if !defined $if;
      my $type = '';
      if($if =~ /^\d+$/) {$type = 'num'}
      elsif($if =~ /^>\s*([+-]?\d+(?:\.\d+)?)$/)  {$type = 'gt'; $if = $1}
      elsif($if =~ /^>=\s*([+-]?\d+(?:\.\d+)?)$/) {$type = 'gte'; $if = $1}
      elsif($if =~ /^<\s*([+-]?\d+(?:\.\d+)?)$/)  {$type = 'lt'; $if = $1}
      elsif($if =~ /^<=\s*([+-]?\d+(?:\.\d+)?)$/) {$type = 'lte'; $if = $1}
      elsif($if =~ /^([+-]?\d+(?:\.\d+)?)\s*-\s*([+-]?\d+(?:\.\d+)?)$/) {
        $type = 'rng'; $if = "$1-$2";
      }
      elsif($if =~ /^'(.+)'$/) {$type = 'str'; $if = $1}
      elsif($if =~ /^"(.+)"$/) {$type = 'reg'; $if = $1}
      elsif($if =~ /^default$/i) {$default = $then; next}
      
      do_log("Could not determine case type for string '$if'", 0)
         and return if !defined $if;
      $cases{++$case_num}{'if'} = $if;
      $cases{$case_num}{'type'} = $type;
      $cases{$case_num}{'then'} = $then;
    }
    my @case_nums = sort {$a <=> $b} keys %cases;

   # Make sure we have a default value
    $default = 'Unknown' if !defined $default;

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)

    if($oids->{$dep_oid}{'repeat'}) {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$dep_oid}{'repeat'};

     # Make sure we have leaf data for our dependent oid
      if(!defined $oids->{$dep_oid}{'val'}) {
        do_log("Missing repeater data for trans_tswitch on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      for my $leaf (keys %{$oids->{$dep_oid}{'val'}}) {
        my $val = $oids->{$dep_oid}{'val'}{$leaf};

        my $num;
        for my $case (@case_nums) {
          my $if   = $cases{$case}{'if'};
          my $type = $cases{$case}{'type'};
          if($type eq 'num') { $num = $case and last if $val == $if; next}
          if($type eq 'gt')  { $num = $case and last if $val > $if; next}
          if($type eq 'gte') { $num = $case and last if $val >= $if; next}
          if($type eq 'lt')  { $num = $case and last if $val < $if; next}
          if($type eq 'lte') { $num = $case and last if $val <= $if; next}
          if($type eq 'rng') { 
            my ($ll,$ul) = split /-/, $if;
            $num = $case and last if $val >= $ll and $val <= $ul; 
            next;
          }
          if($type eq 'str') { $num = $case and last if $val eq $if; next}
          if($type eq 'reg') { $num = $case and last if $val =~ /$if/; next}
        }
          
        my $then;
        if(defined $num) { $then = $cases{$num}{'then'} }
        else { $then = $default }
        while($then =~ /\{(\S+?)\}/) {
          my $then_oid = $1;
          my $then_oid_val;
          if($oids->{$oid}{'repeat'}) {
            $then_oid_val = $oids->{$then_oid}{'val'}{$leaf};
            if(!defined $oids->{$oid}{'color'}{$leaf}) {
              $oids->{$oid}{'color'}{$leaf} = 
                $oids->{$then_oid}{'color'}{$leaf};
              $oids->{$oid}{'msg'}{$leaf} = $oids->{$then_oid}{'msg'}{$leaf};
            }
          }
          else {$then_oid_val = $oids->{$oid}{'val'}}
          if(!defined $then_oid_val) {
            do_log("Missing repeater data for trans_tswitch on $oid", 0);
            $then_oid_val = 'Undefined';
            if(!defined $oids->{$oid}{'color'}{$leaf}) {
              $oids->{$oid}{'color'}{$leaf} = $oids->{$then_oid}{'color'};
              $oids->{$oid}{'msg'}{$leaf} = $oids->{$then_oid}{'msg'};
            }
          }
          $then =~ s/\{$then_oid\}/$then_oid_val/g;
        }
          
        $oids->{$oid}{'val'}{$leaf}   = $then;
        $oids->{$oid}{'time'}{$leaf}  = time;

      }

      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {
      my $val = $oids->{$dep_oid}{'val'};

     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

      if(!defined $val) {
        do_log("No snmp data found for $dep_oid on $device", 0);
        $oids->{$oid}{'val'}   = "No SNMP data ";
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'clear';
        return;
      }

      my $num;
      for my $case (@case_nums) {
	my $if   = $cases{$case}{'if'};
	my $type = $cases{$case}{'type'};
	if($type eq 'num') { $num = $case and last if $val == $if; next}
	if($type eq 'gt')  { $num = $case and last if $val > $if; next}
	if($type eq 'gte') { $num = $case and last if $val >= $if; next}
	if($type eq 'lt')  { $num = $case and last if $val < $if; next}
	if($type eq 'lte') { $num = $case and last if $val <= $if; next}
	if($type eq 'rng') { 
	  my ($ll,$ul) = split /-/, $if;
	  $num = $case and last if $val >= $ll and $val <= $ul; 
	  next;
	}
	if($type eq 'str') { $num = $case and last if $val eq $if; next}
	if($type eq 'reg') { $num = $case and last if $val =~ /$if/; next}
      }
	
      my $then;
      if(defined $num) { $then = $cases{$num}{'then'} }
      else { $then = $default }
      while($then =~ /\{(\S+?)\}/) {
	my $then_oid = $1;
	my $then_oid_val;
	if($oids->{$then_oid}{'repeat'}) {
          do_log("Cant switch to a repeater OID when using a non-repeater" .
            "source OID for trans_tswitch on $oid", 0);
          $then_oid_val = 'Undefined';
	}
	else {
          $then_oid_val = $oids->{$then_oid}{'val'};
	  if(!defined $oids->{$oid}{'color'}) {
	    $oids->{$oid}{'color'} = $oids->{$then_oid}{'color'};
	    $oids->{$oid}{'msg'} = $oids->{$then_oid}{'msg'};
	  }
        }

	if(!defined $then_oid_val) {
	  do_log("Missing repeater data for trans_tswitch on $oid", 0);
	  $then_oid_val = 'Undefined';
	}
	$then =~ s/\{$then_oid\}/$then_oid_val/g;
      }
	
      $oids->{$oid}{'val'}   = $then;
      $oids->{$oid}{'time'}  = time;
      
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }

 # Do regular subsition translations
  sub trans_regsub {
    my ($device, $oids, $oid, $thr) = @_;

    my $data   = $oids->{$oid}{'trans_data'};
    my $repeat = 0;
    my $valid_non_reps = 1;
    my ($invalid_val, $invalid_time, $invalid_color);

   # Rip our dependencies out of the template text
    my ($primary_oid, $expression) = ($1, $2)
      if $data =~ /^\{\s*(\S+?)\s*\}\s*(\/.+\/.*\/)\s*$/;

   # We should always have a dep_oid, due to checks made at template load
   # But just in case...
    do_log("Bad transform data for $primary_oid on $device", 0) and return
      if !defined $primary_oid or !defined $expression;
 
    do_log("$primary_oid not defined on $device for REGSUB transform", 0) 
      and return if !defined $oids->{$primary_oid};

   # Repeat if our suboid is a repeater
    $repeat = 1 if $oids->{$primary_oid}{'repeat'};

   # Go through the expression and replace all non-repeater oids first
    for my $dep_oid ($expression =~ /\{(.+?)\}/g) {
     # Skip non-reps, but note that we have at least one
      if ($oids->{$dep_oid}{'repeat'}) {$repeat = 1; next} 

     # Check that this non-repeater is valid
      my $is_valid = validate_dep($device, $oids, $oid, $dep_oid, 0);

     # If it isnt valid, copy our invalid values so we can assign them
     # To our leaves if we have repeating dependencies
      if($is_valid != 1) {
        $invalid_val   = $oids->{$oid}{'val'};
        $invalid_time  = $oids->{$oid}{'time'};
        $invalid_color = $oids->{$oid}{'color'};
        $valid_non_reps = $is_valid;
        next;
      }

     # Replace non-repeaters with their corresponding value
      my $val = $oids->{$dep_oid}{'val'};
      $expression =~ s/\{$dep_oid\}/$val/g;
    }

   # See if we are a repeating variable type datum
   # (such as that generated by snmpwalking a table)
    if($repeat) {
      my $valid_non_reps = 1;

     # Set our repeater type
      $oids->{$oid}{'repeat'} = $oids->{$primary_oid}{'repeat'};

     # Make sure our primary oid exists
      if(!defined $primary_oid) {
        do_log("Can't find primary oid for trans_regsub on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Cant determine primary oid";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }
      
     # Make sure we have leaf data for our primary oid
      elsif(!defined $oids->{$primary_oid}{'val'}) {
        do_log("Missing primary oid data for trans_regsub on $oid", 0);
        $oids->{$oid}{'val'}{1}   = "Missing primary oid data";
        $oids->{$oid}{'time'}{1}  = time;
        $oids->{$oid}{'color'}{1} = 'red';
        return;
      }

      REGSUBLEAF: for my $leaf (keys %{$oids->{$primary_oid}{'val'}}) {
       # If one or more of our non-repeaters was invalid, all leaves are too
        if($valid_non_reps != 1) {
          $oids->{$oid}{'val'}{$leaf}   = $invalid_val;
          $oids->{$oid}{'time'}{$leaf}  = $invalid_time;
          $oids->{$oid}{'color'}{$leaf} = $invalid_color;
          next;
        }

       # Copy our expression, then replace it
        my $leaf_exp = $expression;
        for my $dep_oid ($expression =~ /\{(.+?)\}/g) {
          next REGSUBLEAF 
            if validate_dep($device, $oids, $oid, $dep_oid, $leaf) != 1;

          my $val = $oids->{$dep_oid}{'val'}{$leaf};
          $leaf_exp =~ s/\{$dep_oid\}/$val/g;
        }

        my $exp_val = $oids->{$primary_oid}{'val'}{$leaf};
        my $eval_cmd = "\$exp_val =~ s$expression";
        my $result = eval($eval_cmd);
        if($@) {
          do_log("Failed eval for REGSUB transform on leaf $leaf of " .
                 "$oid on $device ($@)", 0);
          $oids->{$oid}{'val'}{$leaf}   = 'Failed eval';
          $oids->{$oid}{'time'}{$leaf}  = time;
          $oids->{$oid}{'color'}{$leaf} = 'yellow';
          next REGSUBLEAF;
        }

        $oids->{$oid}{'val'}{$leaf}  = $exp_val;
        $oids->{$oid}{'time'}{$leaf} = time;
      }

     # Now apply our threshold to this data
      apply_thresh_rep($oids, $thr, $oid);
    }

   # Otherwise we are a single entry datum
    else {

     # Set our repeater type
      $oids->{$oid}{'repeat'} = 0;

     # Make sure our dependencies if any are valid
      if($valid_non_reps != 1) {
        $oids->{$oid}{'val'}   = $invalid_val;
        $oids->{$oid}{'time'}  = $invalid_time;
        $oids->{$oid}{'color'} = $invalid_color;
        next;
      }

     # All of our non-reps were substituted earlier, so we can just eval
      my $result = eval($expression);
      if($@) {
        do_log("Failed eval for MATH transform on $oid on $device ($@)", 0);
        $oids->{$oid}{'val'}   = 'Failed eval';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'yellow';
        return;
      }
      $oids->{$oid}{'val'}  = $result;
      $oids->{$oid}{'time'} = time;
        
     # Now apply our threshold to this data
      apply_thresh($oids, $thr, $oid);
    }

    return;
  }



 # Now create our message that we are going to send out
  sub render_msg {
    my ($device, $tmpl, $test, $oids) = @_;

   # Hash shortcut
    my $msg_template = $tmpl->{'msg'};
    my $dev          = \%{$g{'dev_data'}{$device}};

   # No message template?
    if(!defined $msg_template) {
      my $bb_host = $device;
      $bb_host =~ s/\./,/g;

     return "status $bb_host.$test clear " . (scalar localtime) .
        "\n\nCould not locate template for this device.\n"     .
        "Please check devmon logs.\n\n"                         .
        "Devmon version $g{'version'} running on $g{'nodename'}\n";
    }


   # Our outbound message
    my $msg         = '';
    my $rrddata     = '';
    my $curprim     = '';
    my $errors      = '';
    my $worst_color = 'green';
    my $table       = undef;
    my %t_opts;

   # Go through message template line by line
    MSG_LINE: for my $line (split /\n/, $msg_template) {

     # First see if this is table data 
      if($line =~ /^TABLE:\s*(.*)/o) {
        %t_opts = ();
        my $opts = $1;
        if(defined $opts) {
          for my $optval (split /\s*,\s*/, lc $opts) {
            my ($opt,$val) = split /\s*=\s*/, $optval;
            $val = 1 if !defined $val;
            $t_opts{$opt} = $val;
          }
        }

       # Add our html header if necessary
        if(defined $t_opts{'nonhtml'}) {
          $table = '';
        }
        else {
          my $border = (defined $t_opts{'border'}) ? $t_opts{'border'} : 1;
          my $pad    = (defined $t_opts{'pad'}) ? $t_opts{'pad'} : 5;
          $table = "<table border=$border cellpadding=$pad>\n";
        }
        next;
      }

     # If we have seen a TABLE: placeholder, do table logic
      if(defined $table) {

       # First check and see if this is our table header
        if($line !~ /\{(.+?)}/) {

         # Format the line accordingly
          if(defined $t_opts{'nonhtml'}) {
            $table .= "$line\n";  
            $line =~ s/\|/:/g;
          }
          else {
            $line =~ s/\|/<\/td><td>/g;
            $table .= "<tr><td>$line</td></tr>\n";  
          }
          next MSG_LINE;
        }

       # Otherwise its a table data line, we have to parse the oids
        my $alarm_ints = '';

       # Replace our separaters with html
        if(defined $t_opts{'nonhtml'}) {$line =~ s/\|/:/g}
        else                           {$line =~ s/\|/<\/td><td>/g}

       # Make the first oid (from left to right) the primary one
        my $pri = $1 if $line =~ /\{(.+?)}/;
        do_log("Table missing repeater data in $test test for $device", 0)
          and next if !defined $pri;

       # Remove any flags the primary oid might have on it...
        $pri =~ s/\..*//;

       # Make sure we have leaf data for our primary oid
        if(!defined $oids->{$pri}{'val'}) {
          do_log("Missing repeater data for $pri for $test msg", 0);
          next;
        }

       # Make sure our primary OID is a repeater
        if(!$oids->{$pri}{'repeat'}) {
          do_log("Primary OID $pri in $test table is a non-repeater", 0);
          next;
        }


       # If the primary oids leaves are non-numeric, then we cant sort it
       # numerically, we'll have to resort to a cmp
        my @table_leaves = ();
        if($oids->{$pri}{'repeat'} == 2) {
          @table_leaves = keys %{$oids->{$pri}{'val'}};
        }
       # Otherwise sort them numerically ascending
        else {
          @table_leaves = sort {$a <=> $b} keys %{$oids->{$pri}{'val'}};
        }

       # Now go through all oid vals, using the primary's leaves
        T_LEAF: for my $leaf (@table_leaves) {

          my $row_data = $line;
          my $leafrrd  = '';

         # Do some alarm logic
          my $pri_val = $oids->{$pri}{'val'}{$leaf};
          my $alarm = 1; # Alarm by default
          my $a_val  = $dev->{'except'}{$test}{$pri}{'alarm'}     ||
            $dev->{'except'}{'all'}{$pri}{'alarm'}                ||
            $tmpl->{'oids'}{$pri}{'except'}{'alarm'};
          $alarm = ($pri_val =~ /^($a_val)$/) ? 1 : 0  if defined $a_val;

          my $na_val = $dev->{'except'}{$test}{$pri}{'noalarm'}   ||
            $dev->{'except'}{'all'}{$pri}{'noalarm'}              ||
            $tmpl->{'oids'}{$pri}{'except'}{'noalarm'};
          $alarm = 0 if defined $na_val and $pri_val =~ /^($na_val)$/;

         # Now go through all the oids in our table row and replace them
          for my $root ($row_data =~ /\{(.+?)}/g) {

           # Chop off any flags and store them for later
            my $oid  = $root;
            my $flag = $1 if $oid =~ s/\.(.+)$//;

           # Get our oid vars
            my $val   = $oids->{$oid}{'val'}{$leaf};
            my $color = $oids->{$oid}{'color'}{$leaf};
            $val      = 'Undefined' if !defined $val;
            $color    = 'clear' if !defined $color;

           # Check the exception types, if it is an 'ignore'
           # dont include this leaf row if the data for this
           # oid matches, if it is an 'only' type, ONLY include
           # this leaf row if the data matches
            my $ignore = $dev->{'except'}{$test}{$oid}{'ignore'} ||
              $dev->{'except'}{'all'}{$oid}{'ignore'}            ||
              $tmpl->{'oids'}{$oid}{'except'}{'ignore'};
            my $only = $dev->{'except'}{$test}{$oid}{'only'}     ||
              $dev->{'except'}{'all'}{$oid}{'only'}              ||
              $tmpl->{'oids'}{$oid}{'except'}{'only'};
            next T_LEAF if defined $ignore and $val =~ /^($ignore)$/;
            next T_LEAF if defined $only and $val !~ /^($only)$/;


           # If we arent alarming on a value, its green by default
            $color = 'green' if !$alarm;

           # If this is our primary oid, and we are have an alarm
           # variable defined, add this to our alarm header if we are
           # alarming on it; Wrap our header at 60 chars
            if(!defined $t_opts{'noalarmsmsg'} 
               and $oid eq $pri and $alarm) {
              $alarm_ints =~ s/(.{60,}),$/$1)\nAlarming on (/; 
              $alarm_ints .= "$val,";
              $curprim = $val;
            }
              
           # See if we have a valid flag, if so, replace the
           # place holder with flag data, if not, just replace
           # it with the oid value.  Also modify the global color
           # Display a bb/hobbit color string (i.e. "&red ")
            if(defined $flag) {
              if($flag eq 'color') {

               # Honor the 'alarm' exceptions
                $row_data =~ s/\{$root}/&$color /;

               # If this test has a worse color, use it for the global color
                $worst_color = $color if !defined $worst_color or 
                  $colors{$worst_color} < $colors{$color};
              }

             # Display threshold messages if we get the msg flag
              elsif ($flag eq 'msg') {
                my $substr = $oids->{$oid}{'msg'}{$leaf};
                $substr = 'Undefined' if !defined $substr;
                $row_data =~ s/\{$root}/$substr/;
              }

             # Record RRD data
              elsif ($flag eq 'rrd') {
                $row_data =~ s/\{$root}//;
                my $val = $oids->{$oid}{'val'}{$leaf};

               # Only display data on alarming interfaces
                if($alarm) {
                 # Only display numeric oid data
                  if(defined $val and $val =~ /^[+-]?\d+(\.\d+)?$/) {
                    $leafrrd .= "$oid=$val ";
                  }
                  elsif(!defined $val or $val ne 'wait') {
                    do_log("Cant graph non-numeric rrd data for $oid.$leaf" .
                      " on $device",0);
                  }
                }
              }

             # This flag only causes errors (with the color) to be displayed
             # Will also modify global color type
              elsif ($flag eq 'errors') {
                $row_data =~ s/\{$root}//;

                next if $color eq 'green';

               # Get oid msg and replace any inline oid dependencies
                my $oid_msg = $oids->{$oid}{'msg'}{$leaf};
                $oid_msg = 'Undefined' if !defined $oid_msg;
                $oid_msg = parse_deps($oids, $oid_msg, $leaf);
               
               # If this test has a worse color, use it for the global color
                $worst_color = $color if !defined $worst_color or 
                  $colors{$worst_color} < $colors{$color};

               # Now add it to our msg                
                $errors .= "&$color $oid_msg\n";
              }
            }

           # Otherwise just display the oid val
            else {
              my $substr = $oids->{$root}{'val'}{$leaf};
              $substr = 'Undefined' if !defined $substr;
              $row_data =~ s/\{$root}/$substr/;
            }

          }

         # Finished with this row (signified by the primary leaf id)
          if(defined $t_opts{'nonhtml'}) {
            $table .= "$row_data\n";
          }
          else {
            $table .= "<tr><td>$row_data</td></tr>\n";
          }

          if($leafrrd ne '') {
            $leafrrd =~ s/\s+$//;
            $curprim =~ s/[\/ ]/_/g;
            $rrddata .= "<!-- RRD : $curprim : $leafrrd -->\n";
          }
        }

        $table .= "</table>\n" if !defined $t_opts{'nonhtml'};

       # If we are display alarms, fix the message up to look nice
        if(!defined $t_opts{'noalarmsmsg'}) {
          if($alarm_ints ne '') {$alarm_ints =~ s/(.+),/Alarming on ($1)\n/s}
          else {$alarm_ints = "Not alarming on any values\n"}
        }

       # Put the alarm ints on bottom if requested to do so
        if(defined $t_opts{'alarmsonbottom'}) {
          $msg = join '', ($msg,$table,$alarm_ints);
        }
        else {
          $msg = join '', ($msg,$alarm_ints,$table);
        }

       # Add rrd data
        $msg = join '', ($msg,"<table>\n$rrddata</table>\n") if $rrddata ne '';

       # Clear our table status
        $table = undef;

      } # End table data processing
                
     # Not table data, so it should be a non-repeater type variable
      else {

        for my $root ($line =~ /\{(.+?)}/g) {

         # Chop off any flags and store them for later
          my $oid  = $root;
          my $flag = $1 if $oid =~ s/\.(.+)$//;
          $flag    = '' if !defined $flag;

         # Get our oid vars
          my $val   = $oids->{$oid}{'val'};
          my $color = $oids->{$oid}{'color'};
          $val      = 'Undefined' if !defined $val;
          $color    = 'clear' if !defined $color;

         # See if we have a valid flag, if so, replace the
         # place holder with flag data, if not, just replace
         # it with the oid value
          if($flag ne '') {
            if($flag eq 'color') {
             # If this test has a worse color, use it for the global color
              $worst_color = $color if !defined $worst_color or 
                $colors{$worst_color} < $colors{$color};
              $line =~ s/\{$root}/\&$color /;
            }

            elsif ($flag eq 'msg') {
              my $data = $oids->{$oid}{'msg'};
              $data = "Undefined" if !defined $data;
              $data = parse_deps($oids, $data, undef);
              $line =~ s/\{$root}/$data/;
            }

           # This flag only causes errors (with the color) to be displayed
           # Can also modifies global color
            elsif ($flag eq 'errors') {
              $line =~ s/\{$root}/#ERRORONLY#/;

             # Skip this value if it is green
              next if !defined $color or $color eq 'green';

             # Get oid msg and replace any inline oid dependencies
              my $oid_msg = $oids->{$oid}{'msg'};
              $oid_msg = parse_deps($oids, $oid_msg, undef);
                 
             # If this test has a worse color, use it for the global color
              $worst_color = $color if !defined $worst_color or 
                $colors{$worst_color} < $colors{$color};

             # Now add it to our msg                
              $errors .= "&$color $oid_msg\n";
            }

          }

         # Graph rrd data
          elsif ($flag eq 'rrd') {
            $line =~ s/\{$root}//;
            my $val = $oids->{$oid}{'val'};
           # Only display numeric oid data
            if(defined $val and $val =~ /^[+-]?\d+\.\d+?$/) {
              $line .= " <!-- RRD : $oid : $val -->";
            }
            elsif(!defined $val or $val ne 'wait') {
              do_log("Cant graph non-numeric rrd data for $oid on $device",0);
            }
          }

          else {
            my $val = $oids->{$oid}{'val'};
            $val = "Unknown" if !defined $val;

            $line =~ s/\{$root}/$val/;
          }

        }

       # Avoid blank error lines
        $line = ($line eq '#ERRORONLY#') ? '' : "$line\n";
        $msg .= $line;
      }

    } 

   # Add our errors
    $msg = join "\n", ($errors, $msg) if $errors ne '';

   # Now add our header so bb/hobbit can determine the page color
    my $bb_host = $device;
    $bb_host =~ s/\./,/g;
    $msg = "status $bb_host.$test $worst_color " . (scalar localtime) .
      "\n\n$msg"; 

   # Add our oh-so-stylish devmon footer
    $msg .= "\n\nDevmon version $g{'version'} " .
      "running on $g{'nodename'}\n";

   # Now add a bit of logic to allow a 'cleartime' window, where a test on
   # a device can be clear for an interval without being reported as such
    if($worst_color eq 'clear') {
      $g{'numclears'}{$device}{$test} = time
        if !defined $g{'numclears'}{$device}{$test};
    }
    else {
     # Clear our clear counter if this message wasnt clear
      delete $g{'numclears'}{$device}{$test}
        if defined $g{'numclears'}{$device} 
        and defined $g{'numclears'}{$device}{$test};
    }

   # Now return null if we are in our 'cleartime' window
    if(defined $g{'numclears'}{$device} and 
       defined $g{'numclears'}{$device}{$test}) {
      my $start_clear = $g{'numclears'}{$device}{$test};
      return if time - $start_clear < $g{'cleartime'};
    }

   # Looks like we are good to return our completed message
    return $msg;
  }


 # Check alarm exceptions to see if a given oid should be
 # alarmed against.  Return true if it should be ignored.
  sub check_alarms {
    my ($tmpl, $dev, $test, $oid, $val) = @_;

    if(defined $dev->{'except'}{$test}{$oid}{'noalarm'}) { 
      my $match = $dev->{'except'}{$test}{$oid}{'noalarm'};
      return 1 if $val =~ /^($match)$/;
    }
    elsif(defined $dev->{'except'}{'all'}{$oid}{'noalarm'}) { 
      my $match = $dev->{'except'}{'all'}{$oid}{'noalarm'};
      return 1 if $val =~ /^($match)$/;
    }
    elsif(defined $tmpl->{'oids'}{$oid}{'except'}{'noalarm'}) { 
      my $match = $tmpl->{'oids'}{$oid}{'except'}{'noalarm'};
      return 1 if $val =~ /^($match)$/;
    }

    if(defined $dev->{'except'}{$test}{$oid}{'alarm'}) { 
      my $match = $dev->{'except'}{$test}{$oid}{'alarm'};
      return 1 if $val !~ /^($match)$/;
    }
    elsif(defined $dev->{'except'}{'all'}{$oid}{'alarm'}) { 
      my $match = $dev->{'except'}{'all'}{$oid}{'alarm'};
      return 1 if $val !~ /^($match)$/;
    }
    elsif(defined $tmpl->{'oids'}{$oid}{'except'}{'alarm'}) { 
      my $match = $tmpl->{'oids'}{$oid}{'except'}{'alarm'};
      return 1 if $val !~ /^($match)$/;
    }
  }

 # For an oid msg, convert all inline oid dependencies to their textual value
 # Watch out for recursion loops here, as this sub can call itself recursively
  sub parse_deps {
    my ($oids, $msg, $leaf, $depth) = @_;
    my $flags = 'color|msg';

   # Try to guard against recursion loops
    $depth = 1 if !defined $depth;
    return "Recursion limit exceeded." if ++$depth > 10;
    

   # Make sure we have a message to parse!
    return "Undefined msg" if !defined $msg;

   # Go through all the oids that we depend on
    for my $dep_oid ($msg =~ /\{(.+?)\}/g) {
       my $oid = $dep_oid;
       my $val;
      # See if our oid has any flags appended
       my $flag = '';
       $flag = $1 if $oid =~ s/\.($flags)$//;

     # Do repeater types if we have a leaf
      if(defined $leaf) {
       # Evaluate flag type
        if($flag eq 'color') {
          $val = "&" . $oids->{$oid}{'color'}{$leaf} . " ";
        }
        elsif ($flag eq 'msg') {
          my $depmsg = $oids->{$oid}{'msg'}{$leaf};
          $val = parse_deps($oids,$depmsg, $leaf, $depth);
        }
        else {
          $val = $oids->{$oid}{'val'}{$leaf};
        }

        if(!defined $val) {
          do_log("Missing msg data for $dep_oid on leaf $leaf", 0);
          $val = 'Undefined';
        } 
      }

     # Not a repeater
      else {
       # Evaluate flag type
        if($flag eq 'color') {
          $val = "&" . $oids->{$oid}{'color'} . " ";
        }
        elsif ($flag eq 'msg') {
          my $depmsg = $oids->{$oid}{'msg'};
          $val = parse_deps($oids,$depmsg, undef, $depth);
        }
        else {
          $val = $oids->{$oid}{'val'};
        }
        if(!defined $val) {
          do_log("Missing msg data for $dep_oid", 0);
          $val = 'Undefined';
        } 
      }
          
      $msg =~ s/\{$dep_oid}/$val/g;
    }

    return $msg;
  }
      
    
 # Apply thresholds to a supplied non-repeater oid, save in the oids hash
  sub apply_thresh {
    my ($oids, $thr, $oid) = @_;

    my $oid_val = $oids->{$oid}{'val'};
    if($oid_val eq 'wait') {
      $oids->{$oid}{'msg'} = 'wait';
      $oids->{$oid}{'color'} = 'green';
      next;
    }

   # Sneakily sort our color array by color severity
    for my $color (@color_order) {
      my $threshes;

     # Determine if we use custom or template thresholds
      if(defined $thr->{$oid}{$color}) {
        $threshes = $thr->{$oid}{$color};
      }
      elsif(defined $oids->{$oid}{'thresh'}{$color}) {
        $threshes = $oids->{$oid}{'thresh'}{$color}{'val'};
      }

      next if !defined $threshes;

      my $match = 0;
     # Split our comma-delimited thresholds up
      for my $thresh (split /\s*,\s*/, $threshes) {
	$match = 1;

       # Determine if our thresholds are numeric or string based
	if($thresh =~ /^(>|<|>=|<=|=)?\s*([+-]?\d+(\.\d+)?)$/) {
	  my ($op, $limit) = ($1, $2);
	  $op = '>' if !defined $op;
	  last if $op eq '>'  and $oid_val > $limit;
	  last if $op eq '<'  and $oid_val < $limit;
	  last if $op eq '>=' and $oid_val >= $limit;
	  last if $op eq '<=' and $oid_val <= $limit;
	  last if $op eq '='  and $oid_val == $limit;
	}

       # Okay, its not numeric, it must be string based
	else {
	 # Do our automatching for blank thresholds
	  last if $thresh eq '_AUTOMATCH_';
	  last if $oid_val =~ /$thresh/;
	}

	$match = 0;
      }

      if($match) {
	$oids->{$oid}{'color'} = $color;
	$oids->{$oid}{'msg'}   = $oids->{$oid}{'thresh'}{$color}{'msg'};
	return;
      }
    }

   # Check to make sure to assign a color by default
    $oids->{$oid}{'color'} = 'green';
  }
          
 # Apply thresholds to a supplied repeater oid, save in the oids hash
  sub apply_thresh_rep {
    my ($oids, $thr, $oid) = @_;

    # Sneakily sort our color array by color severity
    APTHRLEAF: for my $leaf (keys %{$oids->{$oid}{'val'}}) {
      my $oid_val = $oids->{$oid}{'val'}{$leaf};

      if($oid_val eq 'wait') {
        $oids->{$oid}{'msg'}{$leaf} = 'wait';
        $oids->{$oid}{'color'}{$leaf} = 'green';
        next;
      }

      for my $color (@color_order) {
        my $threshes;

       # Determine if we use custom or template thresholds
        if(defined $thr->{$oid}{$color}) {
          $threshes = $thr->{$oid}{$color};
        }
        elsif(defined $oids->{$oid}{'thresh'}{$color}) {
          $threshes = $oids->{$oid}{'thresh'}{$color}{'val'};
        }

        next if !defined $threshes;
        my $match = 0;

       # Split our comma-delimited thresholds up
	for my $thresh (split /\s*,\s*/, $threshes) {
	  $match = 1;
	 # Determine if our thresholds are numeric or string based
	  if($thresh =~ /^(>|<|>=|<=|=)?\s*([+-]?\d+(\.\d+)?)$/) {
	    my ($op, $limit) = ($1, $2);
	    $op = '>' if !defined $op;
	    last if $op eq '>'  and $oid_val > $limit;
	    last if $op eq '<'  and $oid_val < $limit;
	    last if $op eq '>=' and $oid_val >= $limit;
	    last if $op eq '<=' and $oid_val <= $limit;
	    last if $op eq '='  and $oid_val == $limit;
	  }

	 # Okay, its not numeric, it must be string based
	  else {
	   # Do our automatching for blank thresholds
	    last if $thresh eq '_AUTOMATCH_';
	    last if $oid_val =~ /$thresh/;
	  }

	  $match = 0;
	}

	if($match) {
	  $oids->{$oid}{'color'}{$leaf} = $color;
	  $oids->{$oid}{'msg'}{$leaf} = $oids->{$oid}{'thresh'}{$color}{'msg'}; 
	  next APTHRLEAF;
	}
      }

     # Check to make sure to assign a color by default
      $oids->{$oid}{'color'}{$leaf} = 'green';
    }

    return;
  }

 # Convert # of seconds into an elapsed string
  sub elapsed {
    my ($secs) = @_;

    return $secs if $secs !~ /^\d+(\.\d+)?$/;

    my $days  = int ($secs / 86400);    $secs -= ($days  * 86400);
    my $hours = int ($secs / 3600);     $secs -= ($hours * 3600);
    my $mins  = int ($secs / 60);       $secs -= ($mins  * 60);

    my $out = sprintf "%-2.2d:%-2.2d:%-2.2d", 
      $hours, $mins, $secs; 
    $out = sprintf "%d day%s, %s", $days, 
      ($days == 1) ? "" : "s", $out if $days;

    return $out;
  }

 # Convert # of seconds into an elapsed string
  sub date {
    my ($secs) = @_;

    my ($sec, $min, $hour, $day, $mon, $year) = localtime($secs);
    $year += 1900; ++$mon;

    my $out = sprintf "%d-%-2.2d-%-2.2d, %-2.2d:%-2.2d:%-2.2d",
      $year, $mon, $day, $hour, $min, $sec;
    return $out;
  }


 # Validate oid dependencies
  sub validate_dep {
    my ($device, $oids, $oid, $dep_oid, $leaf, $regex) = @_;

    if(defined $leaf and $leaf != 0) {
      my $val = $oids->{$dep_oid}{'val'}{$leaf};
      if(!defined $val) {
        if($oids->{$dep_oid}{'transform'}) {
          do_log("Cant find translated data for leaf $leaf " .
                 "of $dep_oid on $device", 0);
        }
        else {
          do_log("Cant find snmp data for leaf $leaf " .
                 "of $dep_oid on $device", 0)
        }
        $oids->{$oid}{'val'}{$leaf}   = 'No data';
        $oids->{$oid}{'time'}{$leaf}  = time;
        $oids->{$oid}{'color'}{$leaf} = 'clear';
        return 0;
      }
      elsif($oids->{$dep_oid}{'val'}{$leaf} eq 'wait') {
        $oids->{$oid}{'val'}{$leaf}   = 'wait';
        $oids->{$oid}{'time'}{$leaf}  = time;
        $oids->{$oid}{'color'}{$leaf} = 'green';
        return 2;
      }
      elsif(defined $regex and $regex ne '' and $val !~ /$regex/) {
        do_log("Data type mismatch for leaf $leaf " .
                 "of $dep_oid on $device: '$val' does not match " .
                 "regex '$regex'", 0);
        my $val = $oids->{$dep_oid}{'val'}{$leaf};
        $oids->{$oid}{'val'}{$leaf}   = 'Bad data';
        $oids->{$oid}{'time'}{$leaf}  = time;
        $oids->{$oid}{'color'}{$leaf} = 'clear';
        return 0;
      }
    }

   # Non repeater
    else {
      my $val = $oids->{$dep_oid}{'val'};
      if(!defined $val) {
        if($oids->{$dep_oid}{'transform'}) {
          do_log("Cant find translated data for $dep_oid on $device", 0);
        }
        else {
          do_log("Cant find snmp data for $dep_oid on $device", 0)
        }
 
        $oids->{$oid}{'val'}   = 'No data';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'clear';
        return 0;
      }
      elsif($oids->{$dep_oid}{'val'} eq 'wait') {
        $oids->{$oid}{'val'}   = 'wait';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'green';
        return 2;
      }
      elsif(defined $regex and $regex ne '' and $val !~ /$regex/) {
        do_log("Data type mismatch for " .
                 "$dep_oid on $device: '$val' does not match " .
                 "regex '$regex'", 0);
        $oids->{$oid}{'val'}   = 'Bad data';
        $oids->{$oid}{'time'}  = time;
        $oids->{$oid}{'color'} = 'clear';
        return 0;
      }
    }
  
    return 1;
  } 

1;
